#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include "types.h"
#include "points.h"
#include "objects.h"
#include "rand.h"

#define pi  3.141592653589793238462643

typedef struct 
{
  double x,y,z;
  double bii, decl;
} galaxy;

/* declarations */
void cluster(double**, parlist*);

void read_cluster(galaxy*,int*,parlist*);
void cl_select_hemisphere(galaxy*,int*,parlist*);
void cl_subsample(galaxy*, int, int);
void cl_vollimit(galaxy *, int*);
boolean cl_read_line(FILE*, char*);
void cl_flip_gal(galaxy*, galaxy*);
char* int_to_string(int);
char* make_fname(int, parlist*);

/* --------------------------------------------------------------- */
/*                                                                 */
/*   Reads the information from the cluster catalog file           */
/* --------------------------------------------------------------- */
void read_cluster(galaxy*gal, int*n_obj, parlist*par)
{
  FILE  *fp;                    /* file pointer  */
  int  i, j;
  char str1[256], str2[256];
  char *fname;
  static int count = 1;

  /* do the numbering */
  fname = make_fname(count, par);
  count++;

  fp = fopen(fname,"r");             /* open file */
  if (fp == NULL) {
    printf("Open file error in read_cluster");
    return;
  }
  fprintf(par->outfp, "# using file %s\n", fname);

  /* read the coordinates velocities and flux */
  j=0;
  while( cl_read_line(fp, str1) ){
    for (i=0; i<5; i++){
      if (i == 0)
	strcpy(str2,strtok(str1," \t")); /* read first token */
      else
	strcpy(str2,strtok(NULL," \t")); /* read next token  */
      
      if ((str2 == NULL) || (strcmp(str2,"\n") == 0)) {
	fprintf(stderr,"Line read-error in read_points : too few columns !");
	fclose(fp);
	return;
      }
      
      switch(i) {
      case 0 : gal[j].x    = strtod(str2, (char**)NULL);
      case 1 : gal[j].y    = strtod(str2, (char**)NULL);
      case 2 : gal[j].z    = strtod(str2, (char**)NULL);
      case 3 : gal[j].bii  = strtod(str2, (char**)NULL);
      case 4 : gal[j].decl = strtod(str2, (char**)NULL);
      }
    }
    /* increment counter */
    j++;
  }  
  /* number of read in clusters */
  *n_obj = j;

  fclose(fp);

  return; 
}

/* ------------------------ */
/*   the cluster catalog    */
/*                          */
/*   may alter par->n       */
/* ------------------------ */
void cluster(double**p, parlist*par)
{
  int    i, j;
  int    n_gal, n_used;
  galaxy *gal;
  double max_vel, vel;
  double sim_scale, sim_offset, phys_scale;
  double tmp;
  static boolean first=true;
  static int mean_num = 0, count = 1;

  if(par->hemisphere != 1 && par->hemisphere != 2) {
    fprintf(stderr, "use option -H1 or -H2 to select hemisphere\n");
    exit(0);
  }

  /* maximum number of clusters in the file */
  n_gal = 600;

  /* simulation data is in the range [-240, 240] */
  sim_scale  = 480.;
  sim_offset = 240.;

  /* the scaling of [0,1] to [0, 500 h^-1 Mpc] */
  phys_scale = 480.;

  /* write comments to output file */
  if(first) fprintf(par->outfp, "# Comments on clusters \n#\n");

  /* allocate memory and initialize */
  gal = (galaxy*)calloc(n_gal, sizeof(galaxy));

  /* read in the galaxies from par.pointfile */
  read_cluster(gal, &n_gal, par);
  if(first) fprintf(par->outfp, "#  %d cluster in the input file \n",
		    n_gal);

  /* read in the galaxies from par.pointfile */
  cl_vollimit(gal, &n_gal);
  if(first) fprintf(par->outfp, "#  %d clusters with r <= 240 Mpc \n",
		    n_gal);

  /* calculate max_vel */
  for(i=0, max_vel = 0; i<n_gal; i++) {
    vel = sqrt(gal[i].x*gal[i].x + gal[i].y*gal[i].y + gal[i].z*gal[i].z);
    if(vel > max_vel) max_vel = vel;
  }
  if(first) {
    fprintf(par->outfp, "#  maximum radial distance = %g\n", max_vel);
  }

  /* select in hemisphere */
  cl_select_hemisphere(gal, &n_gal, par);
  if(first) fprintf(par->outfp, 
		    "#\n#  %d galaxies in %s hemisphere \n", 
		    n_gal,
		    (par->hemisphere==1 ? "northern" 
		     : (par->hemisphere==2 ? "southern" : "none")) );

  /* subsample n_used galaxies from n_gal */
  n_used = par->subsample/100. * n_gal;
  tmp = (par->subsample/100. * (double)n_gal) - (double)n_used;
  if(rand1() > tmp && tmp > 0) n_used++;
  cl_subsample(gal, n_used, n_gal);
  if(first) fprintf(par->outfp, "#  subsampled %d from %d galaxies\n", 
		    n_used, n_gal);

  /* the mean number of galaxies */
  mean_num += n_used;
  if(count > 8) fprintf(par->outfp, "# mean number in sample %g \n#\n", 
			(double)mean_num/(double)count);
  count++;

  if(first) {
    fprintf(par->outfp, 
	    "#\n#  now we rescale\n");
    fprintf(par->outfp, 
	    "#  maximum radial distance within the sample = %g\n", max_vel);
  }

  /* determine the scaling factor */
  par->scale = phys_scale;
  if(first) fprintf(par->outfp, 
		    "#  scaling factor                     = %g\n", 
		    par->scale);

  /* we use onle n_used */
  par->n = n_used;
  
  /* rescale and shift within [0,1]^3 */
  for(i = 0; i < par->n; i++) {
    p[i][0] = (gal[i].x + sim_offset)/sim_scale;
    p[i][1] = (gal[i].y + sim_offset)/sim_scale;
    p[i][2] = (gal[i].z + sim_offset)/sim_scale;
  }

  /* free memory */
  free((void*)gal);

  fprintf(stderr, "sampled %d galaxies from %d\n", par->n, n_gal);

  if(first) {
    fprintf(par->outfp, "#\n");
    fflush(par->outfp);
  }

  first = false;
  return;
}

/* -------------------------------------------------- */
/*                                                    */
/* select galaxies in northern or southern hemisphere */
/* shuffle the galaxies to the n_gal first positions  */
/* in gal                                             */
/* -------------------------------------------------- */
void cl_select_hemisphere(galaxy *gal, int*n_gal, parlist *par)
{
  int i, j=0;

  for(i=0; i<(*n_gal); i++)
    switch(par->hemisphere) {
    case 1 :                       /* northern */
      if(gal[i].bii > 30*pi/180) {
	cl_flip_gal(&gal[j], &gal[i]);
	j++;
      }
      break;
    case 2 :                       /* southern */
      if(gal[i].bii < -30*pi/180) {
	cl_flip_gal(&gal[j], &gal[i]);
	j++;
      }
      break;
    default :                      /* no zoa; complete sphere */
      j = par->n;
      break;	
    }
  
  (*n_gal) = j;
  return;
}

/* -------------------------------------------- */
/*                                              */
/* subsample n_used galaxies from the first     */
/* n_gal galaxies, shuffle the galaxies to the  */
/* n_used first positions in gal                */
/* -------------------------------------------- */
void cl_subsample(galaxy *gal, int n_used, int n_hem)
{
  int i, irand;

  /* nothing to do */
  if(n_used == n_hem) return;

  /* draw the galaxies */
  for(i=0; i<n_used; i++) {
    irand = i + (int)((n_hem - i)* rand1() - 10e-13);
    cl_flip_gal(&gal[i],&gal[irand]);
  }

  return;
}

/* -------------------------------------------- */
/*                                              */
/* select only clusters with r <= 240           */
/* -------------------------------------------- */
void cl_vollimit(galaxy *gal, int*n_gal)
{
  int i, n;
  double r;

  /* check the clusters */
  for(i=0, n=0; i<(*n_gal); i++) {
    r = sqrt(gal[i].x*gal[i].x + gal[i].y*gal[i].y + gal[i].z*gal[i].z);
    if(r <= 240) {
      cl_flip_gal(&gal[i],&gal[n]);
      n++;
    }
  }

  (*n_gal) = n;
  return;
}

/* flips two galaxies */
void cl_flip_gal(galaxy *g1, galaxy *g2)
{
  galaxy g;

  g.x = g2->x; g.y = g2->y; g.z = g2->z;
  g.bii = g2->bii; g.decl = g2->decl;

  g2->x = g1->x; g2->y = g1->y; g2->z = g1->z;
  g2->bii = g1->bii; g2->decl = g1->decl;

  g1->x = g.x; g1->y = g.y; g1->z = g.z;
  g1->bii = g.bii; g1->decl = g.decl;
}

/*-----------------------------------------------------------------*/
/*   reads in one line from a file, ignores commented lines        */
/*   and lines containing only "\n"                                */
/*                                                                 */
/*   Parameter : fp    : file pointer                              */
/*               str   : pointer to string of minimal length 255   */
/*                                                                 */
/* --------------------------------------------------------------- */
boolean cl_read_line(FILE *fp, char *line)
{
  char *ptr;

  /* loop till one valid line is read in, exit with FALSE when EOF */
  do
    {
                    /* maximum number of characters per row is 255 */
      if ( fgets(line, 255, fp) == NULL){
	/* we reached EOF */
	return false;
      }
                          /* search for comment sign #, throw away */
                          /* the rest of the line                  */
      if ((ptr = strchr(line, '#')) != NULL)
	*ptr = '\0';
                        /* throw away newline \n and anything past */
      if ((ptr = strchr(line, '\n')) != NULL)
	*ptr = '\0';
    }
  while((strlen(line) == 0));
  return true;
}

/* integer to string conversion (simple minded version) */
char* int_to_string(int i)
{
  switch(i)
    {
    case 0: return "0";
    case 1: return "1";
    case 2: return "2";
    case 3: return "3";
    case 4: return "4";
    case 5: return "5";
    case 6: return "6";
    case 7: return "7";
    case 8: return "8";
    case 9: return "9";
    }
}

/* input file name */
char* make_fname(int i, parlist*par)
{
  char *str, *nstr;
  
  str  = (char*)calloc(256,sizeof(char));
  nstr = (char*)calloc(2,sizeof(char));

  nstr = int_to_string(i);
  str  = strcpy(str, par->infile);
  str  = strcat(str, ".");
  str  = strcat(str, nstr);

  return str;
}
