/*
    A program for calculating Minkowski Functionals of the boolean grain
    model in 3D
    Copyright (C) 1996  Jens Schmalzing and Martin Kerscher

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include "types.h"
#include "minkowski.h"
#include "geometry.h"
#include "objects.h"
#include "points.h"
#include "rand.h"
#include "org.h"
#include "recipes.h"

#include "bubble.h"


int main(int nar, char *ar[]) 
{
  int i,j,k, nrad;
  double**store, **points, *vd,*v, rad, brad, num, square;
  pattern all;
  parlist par;
  static boolean first=true;

  /* initialize all the parameters, evaluate the commandline */
  /* and initialize the seeds */
  init_params(&par, ar, nar);

  /* memory for storage */
  nrad = (par.hirad-par.lorad)/par.drad + 1; 
  store = dmatrix(0,8,0,nrad-1);
  for(i=0; i<9; i++) for(j=0; j<nrad; j++) store[i][j] = 0;

  fflush(par.outfp);

  /* loop over samples */
  for(i = 1; i <= par.length; i++) {
    /* points and walls */
    if(strcmp(par.infile, "nofile") == 0) {
      /* initialize the pattern and create window */
      if(first) {
	fprintf(par.outfp, "# using cubic window\n");
	cubewalls(&all);
	first=false;
      }
      /* generate binomial Distribution (almost poisson) within walls */
      patternbinomial(&points, &par, &all);
    } else {
#ifdef BUBBLE
      /* read bubbles */
      bubble(&points, &par); 
      /* initialize the pattern and create window */
      if(first) {
        fprintf(par.outfp, "# using cubic window\n");
        cubewalls(&all);
        first=false;
      }
#endif
    }
    fflush(par.outfp);


    /* loop over radii */
    for(j=0; j<nrad; j++) {
      rad = par.lorad+j*par.drad; fprintf(stderr,"%5d%10f",i,rad);
      store[0][j] = rad;

      /* do we construct a buffer? */
      if(par.rmbound) brad = rad;
      else brad =0.;
      /* create buffer and calculate Minkowski functionals */
      if( (vd = buffer(&all,brad,&par)) ) {
	fprintf(stderr," chi is %10.4f",vd[3]);
	addballs(&all,par.n,points,rad);
	v = minkowski(&all,&par,rad); 
	fprintf(stderr," chi is %10.4f",v[3]);
	if(par.rmbound) {
	  /* deconvolution of the boundary */
	  removeboundary(v,vd);
	} else {
	  /* calculate volume density */
	  for(k=0; k<=3; k++) v[k] /= vd[0];
	}
	for(k=0; k<=3; k++) 
	  store[k+1][j] += v[k], store[k+5][j] += v[k]*v[k];

	/* free memory */
	for(k=0; k<all.n; k++) removebody(&all.b[k]);
	free((void*)v); free((void*)vd);
      }
      fprintf(stderr,"\n"); fflush(stderr);
    }  /* end loop over radii */
  }   /* end loop over samples */



  /* calculate measures of the unbuffered window */
  vd = buffer(&all,0,&par);
  fprintf(par.outfp,"# rescaled measures of the window:\n");
  fprintf(par.outfp,"#  v0 = %g, v1 = %g, v2 = %g, v3 = %g\n",
	  vd[0], vd[1], vd[2], vd[3]);
  fprintf(par.outfp,"# measures of the window:\n");
  fprintf(par.outfp,"#  v0 = %g, v1 = %g, v2 = %g, v3 = %g\n#\n",
	  vd[0]*pow(par.scale,3), vd[1]*pow(par.scale,2), 
	  vd[2]*pow(par.scale,1), vd[3]);


  /* write explanation to output file */
  Explain(par.outfp,ar[0],&par);
  fprintf(par.outfp,"#\n# radius    ");
  if(par.length == 1) {
    for(i = 0; i <= 3; i++)fprintf(par.outfp,"v%d           ",i); 
  } else { 
    for(i = 0; i <= 3; i++)fprintf(par.outfp,"v%d mean      ",i);
  }
  for(i = 0; i <= 3; i++)fprintf(par.outfp,"v%d theory    ",i);
  if(par.length > 1) 
    for(i = 0; i <= 3; i++)fprintf(par.outfp,"v%d error     ",i);
  fprintf(par.outfp,"\n");


  /* output mean and variance of stored values */
  num = (double)par.length;
  for(i = 0; i < nrad; i++) {
    /* theoretical measures */
    v = theory(store[0][i], &par, vd); 
    /* scale radius */
    store[0][i] *= par.scale; 
    /* calculate and scale the measures and errors */
    for(j=0; j<4; j++) {
      if(par.length > 1) {
	square = (store[j+5][i] - pow(store[j+1][i],2)/num)/(num-1);
	store[j+5][i] = safersqrt(square)/pow(par.scale, j);
      }
      store[j+1][i] /= num*pow(par.scale, j);
      /* theory */
      v[j] /= pow(par.scale, j);
    }
    /* radius */
    fprintf(par.outfp,"%10.5f", store[0][i]);
    /* mean measures */
    for(j = 1; j <= 4; j++)fprintf(par.outfp," %12.5g", store[j][i]);
    /* theoretical values, if required */
    fprintf(par.outfp," %12.5g %12.5g %12.5g %12.5g",v[0], v[1],v[2],v[3]);
    /* standard error of measures */
    if(par.length > 1)
      for(j = 1; j <= 4; j++) fprintf(par.outfp," %12.5g", store[j+4][i]);
    fprintf(par.outfp,"\n");
  }
  fclose(par.outfp);


  /* free memory */
  free_dmatrix(store,0,8,0,nrad-1);
  free_dmatrix(points,0,par.n-1,0,3);

  exit(0);
}
