/* ================================================================= */
/* This file contains code copyrighted by Numerical Recipes Software */
/* 1985, 1987 P.O. Box 243, Cambridge, MA 02238                      */
/* ================================================================= */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "types.h"

/* ------------ */
/* declarations */
/* ------------ */
void nrerror(char[]);
double*dvector(int,int);
double**dmatrix(int,int,int,int);
void free_dvector(double*,int,int);
void free_dmatrix(double**,int,int,int,int);

/* ------------------------------------ */
/*                                      */
/* some modified numerical recipes for  */
/* the calculation of integrals         */
/* ------------------------------------ */
double romberg(double(*)(double,section*,body*,pattern*),
	       section*,body*,pattern*,double,double,  char*);
double trapez(double(*)(double,section*,body*,pattern*),
	      section*,body*,pattern*,double,double);
double mytrapzd(double(*)(double,section*,body*,pattern*),
		section*,body*,pattern*,double, double, int);
void polint(double*,double*,int,double,double*,double*);

void nrerror(char error_text[])
{
	void exit();

	fprintf(stderr,"Numerical Recipes run-time error...\n");
	fprintf(stderr,"%s\n",error_text);
	fprintf(stderr,"...now exiting to system...\n");
	exit(1);
}

double *dvector(int nl,int nh)
{
	double *v;

	v=(double *)malloc((unsigned) (nh-nl+1)*sizeof(double));
	if (!v) nrerror("allocation failure in dvector()");
	return v-nl;
}

double **dmatrix(int nrl,int nrh,int ncl,int nch)
{
	int i;
	double **m;

	m=(double **) malloc((unsigned) (nrh-nrl+1)*sizeof(double*));
	if (!m) nrerror("allocation failure 1 in dmatrix()");
	m -= nrl;

	for(i=nrl;i<=nrh;i++) {
		m[i]=(double *) malloc((unsigned) (nch-ncl+1)*sizeof(double));
		if (!m[i]) nrerror("allocation failure 2 in dmatrix()");
		m[i] -= ncl;
	}
	return m;
}

void free_dvector(double* v,int nl,int nh)
{
	free((char*) (v+nl));
}

void free_dmatrix(double** m,int nrl,int nrh,int ncl,int nch)
{
	int i;

	for(i=nrh;i>=nrl;i--) free((char*) (m[i]+ncl));
	free((char*) (m+nrl));
}

/* ================================================================= */


/* for small point sets (<1k) "EPS 1.0e-6"  might be possible        */
/* in large (~100k) strongly clustering point sets 1.0e-6 wont work  */
#define EPS 1.0e-5 

#define JMAX 30
#define JMAXP JMAX+1
#define K 5
double romberg(double(*func)(double,section*,body*,pattern*),
	       section*sect,body*bod,pattern*pat,double a,double b, char *info)
{
  double ss=0,dss=0;
  double s[JMAXP+1],h[JMAXP+1];
  double myss[JMAXP+1],mydss[JMAXP+1];
  int j;

  h[1]=1.0;
  for(j=1;j<=JMAX;j++) {
    s[j]=mytrapzd(func,sect,bod,pat,a,b,j);
    if(j>=K) {
      polint(&h[j-K],&s[j-K],K,0.0,&ss,&dss);
      myss[j]=ss; mydss[j]=dss; 
      if (fabs(dss)<=EPS) return ss;
    }
    s[j+1]=s[j];
    h[j+1]=0.25*h[j];
  }

  /* a crude remedy for triangles with small area: */
  /* writes some info to stderr and return area 0 */
  /* check your results, if you get this error message */
  /* A first remedy you might try to randomly disturb the positions */
  fprintf(stderr, "\nToo many steps in routine QROMB with info %s:\n", info);
  fprintf(stderr, "j %d , value %g , step %g, interp %g, error %g\n", 
	  JMAX, s[JMAX], h[JMAX], myss[JMAX], mydss[JMAX]);
  fprintf(stderr, "returning 0\n");
  
  return 0.;
}
#undef EPS
#undef JMAX
#undef JMAXP
#undef K

#define EPS 1.0e-6
#define JMAX 15
double simpson(double(*func)(double,section*,body*,pattern*),
	       section*sect,body*bod,pattern*pat,double a,double b)
{
  int j;
  double s,st,ost,os;
  
  ost = os =  -1.0e30;
  for (j=1;j<=JMAX;j++) {
    st=mytrapzd(func,sect,bod,pat,a,b,j);
    s=(4.0*st-ost)/3.0;
    if (fabs(s-os) <= EPS*fabs(os)) return s;
    os=s;
    ost=st;
  }
  nrerror("Too many steps in routine QSIMP");
  return 0;
}
#undef EPS
#undef JMAX

#define EPS 1.0e-5
#define JMAX 15
double trapez(double(*func)(double,section*,body*,pattern*),
	      section*sect,body*bod,pattern*pat,double a,double b)
{
  int j;
  double s,olds;

  olds = -1.0e30;
  for (j=1;j<=JMAX;j++) {
    s=mytrapzd(func,sect,bod,pat,a,b,j);
    if (fabs(s-olds) <= EPS*fabs(olds)) return s;
    olds=s;
  }
  nrerror("Too many steps in routine QTRAP");
  return 0;
}
#undef EPS
#undef JMAX


#define FUNC(x) ((*func)(x,sect,bod,pat))
double mytrapzd(double(*func)(double,section*,body*,pattern*),
		section*sect,body*bod,pattern*pat,double a, double b, int n)
{
  double x,tnm,sum,del;
  static double s;
  static int it;
  int j;
  
  if(n==1) {
    it=1;
    return (s=0.5*(b-a)*(FUNC(a)+FUNC(b)));
  } else {
    tnm=it;
    del=(b-a)/tnm;
    x=a+0.5*del;
    for (sum=0.0,j=1;j<=it;j++,x+=del) sum += FUNC(x);
    it *= 2;
    s=0.5*(s+(b-a)*sum/tnm);
    return s;
  }
}
void polint(double*xa, double*ya, int n, double x, double *y, double *dy)
{
  int i,m,ns=1;
  double den,dif,dift,ho,hp,w;
  static double *c=(double*)NULL,*d=(double*)NULL;
  
  if(!c) c=dvector(1,n); if(!d) d=dvector(1,n);
  dif=fabs(x-xa[1]);
  for(i=1;i<=n;i++) {
    if((dift=fabs(x-xa[i]))<dif) ns=i, dif=dift;
    c[i]=d[i]=ya[i];
  }

  *y=ya[ns--];

  for(m=1;m<n;m++) {
    for(i=1;i<=n-m;i++) {
      ho=xa[i]-x;
      hp=xa[i+m]-x;
      w=c[i+1]-d[i];
      if ( (den=ho-hp) == 0.0) nrerror("Error in routine polint");
      den=w/den;
      c[i]=ho*den; d[i]=hp*den;
    }
    *y += (*dy=(2*ns < (n-m) ? c[ns+1] : d[ns--]));
  }
}
